name_fs = "[%(name)s] "
location_fs = "(%(module)s.%(funcName)s:%(lineno)d) "
message_fs = "%(message)s"
default_format_string = f"%(asctime)s.%(msecs)03d [%(levelname)s] {name_fs}{location_fs}{message_fs}"



import logging
import sys
from .filters import DebugInfoFilter, WarningErrorCriticalFilter


def get_stdout_handler(formatter: logging.Formatter) -> logging.StreamHandler:
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.addFilter(DebugInfoFilter())
    stdout_handler.setFormatter(formatter)
    return stdout_handler


def get_stderr_handler(formatter: logging.Formatter) -> logging.StreamHandler:
    stdout_handler = logging.StreamHandler(sys.stderr)
    stdout_handler.setLevel(logging.WARNING)
    stdout_handler.addFilter(WarningErrorCriticalFilter())
    stdout_handler.setFormatter(formatter)
    return stdout_handler



from typing import Literal, Union


class Config:
    logs_dir_path: str = "logs"
    file_name: str = "main.log"
    write_mode: Literal["a", "w"] = "w"
    max_MB: int = 5
    max_count: int = 10
    show_name_label: bool = True
    show_location_label: Union[bool, Literal["auto"]] = "auto"
    format_string: str = None



import logging
from .fs_config import location_fs, message_fs


class ErrorCriticalFormatter(logging.Formatter):
    def format(self, record):
        original_format = str(self._style._fmt)
        if record.levelno >= logging.ERROR and location_fs not in self._style._fmt:
            self._style._fmt = self._style._fmt.split(message_fs)[0] + location_fs + message_fs
        result = super().format(record)
        self._style._fmt = original_format
        return result



from typing import Callable
import logging
import re
from .file_handlers import get_format_string, get_rotating_file_handler
from .stream_handlers import get_stdout_handler, get_stderr_handler
from .formaters import ErrorCriticalFormatter
from .config import Config


class DMLogger:
    config: Config = Config
    __loggers: dict = {}
    __file_handlers: dict = {}

    def __new__(cls, name: str, *args, **kwargs):
        if name not in cls.__loggers:
            cls.__loggers[name] = super().__new__(cls)
        return cls.__loggers[name]

    def __init__(
        self,
        name: str,
        level: str = "DEBUG",
        *,
        write_logs: bool = True,
        print_logs: bool = True,
        file_name: str = None
    ):
        self._logger = logging.getLogger(name)
        level = logging.getLevelName(level.upper())
        self._logger.setLevel(level)
        formatter_instance = ErrorCriticalFormatter if self.config.show_location_label == "auto" else logging.Formatter
        formatter = formatter_instance(get_format_string(self.config), datefmt='%d-%m-%Y %H:%M:%S')

        if write_logs:
            file_name = file_name or self.config.file_name
            if file_name not in self.__file_handlers:
                self.__file_handlers[file_name] = get_rotating_file_handler(file_name, self.config, formatter)
            self._logger.addHandler(self.__file_handlers[file_name])

        if print_logs:
            self._logger.addHandler(get_stdout_handler(formatter))
            self._logger.addHandler(get_stderr_handler(formatter))

    def debug(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.debug, message, **kwargs)

    def info(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.info, message, **kwargs)

    def warning(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.warning, message, **kwargs)

    def error(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.error, message, **kwargs)

    def critical(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.critical, message, **kwargs)

    @staticmethod
    def _log(level_func: Callable, message: any, **kwargs) -> None:
        message = "-- " + str(message) if not (message is None) else ""
        if kwargs:
            dict_string = re.sub(r"'(\w+)':", r"\1:", str(kwargs))
            message = f"{dict_string} {message}"
        level_func(message, stacklevel=3)



import logging


class DebugInfoFilter(logging.Filter):
    def filter(self, record):
        return record.levelno in (logging.DEBUG, logging.INFO)


class WarningErrorCriticalFilter(logging.Filter):
    def filter(self, record):
        return record.levelno in (logging.WARNING, logging.ERROR, logging.CRITICAL)



from logging.handlers import RotatingFileHandler
import logging
import os
from .config import Config
from .fs_config import *


def get_rotating_file_handler(file_name: str, options: Config, formatter: logging.Formatter) -> RotatingFileHandler:
    logs_dir_path = options.logs_dir_path or "logs"
    logs_dir_path = os.path.normpath(logs_dir_path)
    if not os.path.exists(logs_dir_path):
        os.makedirs(logs_dir_path)
    log_path = os.path.join(logs_dir_path, file_name)
    file_handler = RotatingFileHandler(log_path, maxBytes=options.max_MB * 1024 * 1024,
                                       backupCount=options.max_count, encoding="utf-8")
    if options.write_mode == "w" and os.path.exists(log_path):
        file_handler.doRollover()
    file_handler.setFormatter(formatter)
    return file_handler


def get_format_string(options: Config) -> str:
    format_string = options.format_string or default_format_string
    if format_string != default_format_string:
        return format_string
    if options.show_name_label is False:
        format_string = format_string.replace(name_fs, "")
    if options.show_location_label is not True:
        format_string = format_string.replace(location_fs, "")
    return format_string
